# -*- coding: utf-8 -*-
import pytest

import rust_fst.lib as lib
from rust_fst import Map


TEST_ITEMS = [(u"möö", 1), (u"bar", 2), (u"baz", 1337), (u"foo", 2**16)]


def do_build(path=None, keys=TEST_ITEMS, sorted_=True):
    if sorted_:
        it = sorted(keys)
    else:
        it = keys
    if path:
        with Map.build(path) as builder:
            for key, val in it:
                builder.insert(key, val)
    else:
        return Map.from_iter(it)


@pytest.fixture
def fst_map():
    return do_build()


def test_build(tmpdir):
    fst_path = str(tmpdir.join('test.fst'))
    do_build(fst_path)


def test_build_outoforder(tmpdir):
    fst_path = str(tmpdir.join('test.fst'))
    with pytest.raises(lib.TransducerError):
        do_build(fst_path, sorted_=False)


def test_build_baddir():
    fst_path = "/guaranteed-to-not-exist/set.fst"
    with pytest.raises(OSError):
        do_build(fst_path)


def test_build_memory(fst_map):
    assert len(fst_map) == 4


def test_map_contains(fst_map):
    for key, _ in TEST_ITEMS:
        assert key in fst_map


def test_map_items(fst_map):
    items = list(fst_map.items())
    assert items == sorted(TEST_ITEMS)


def test_map_getitem(fst_map):
    for key, val in TEST_ITEMS:
        assert fst_map[key] == val


def test_map_keys(fst_map):
    keys = list(fst_map.keys())
    assert keys == sorted([k for k, _ in TEST_ITEMS])


def test_map_iter(fst_map):
    assert list(fst_map) == sorted([k for k, _ in TEST_ITEMS])


def test_map_values(fst_map):
    values = list(fst_map.values())
    assert values == [v for _, v in sorted(TEST_ITEMS)]


def test_map_search(fst_map):
    matches = list(fst_map.search("bam", 1))
    assert matches == [(u"bar", 2), (u"baz", 1337)]



from cffi import FFI

ffi = FFI()
ffi.set_source('rust_fst._ffi', None)
ffi.cdef("""

    /** ===============================
                   Utility
        =============================== **/

    typedef struct {
        bool  has_error;
        char* error_type;
        char* error_description;
        char* error_display;
        char* error_debug;
    } Context;

    typedef struct BufWriter BufWriter;
    typedef struct Levenshtein Levenshtein;

    Levenshtein* fst_levenshtein_new(Context*, char*, uint32_t);
    void fst_levenshtein_free(Levenshtein*);

    Context* fst_context_new();
    void fst_context_free(Context*);

    void fst_string_free(char*);

    BufWriter* fst_bufwriter_new(Context*, char*);
    void fst_bufwriter_free(BufWriter*);


    /** ===============================
                     Set
        =============================== **/

    typedef struct FileSetBuilder FileSetBuilder;
    typedef struct MemSetBuilder MemSetBuilder;
    typedef struct Set Set;
    typedef struct SetStream SetStream;
    typedef struct SetLevStream SetLevStream;
    typedef struct SetOpBuilder SetOpBuilder;
    typedef struct SetUnion SetUnion;
    typedef struct SetIntersection SetIntersection;
    typedef struct SetDifference SetDifference;
    typedef struct SetSymmetricDifference SetSymmetricDifference;

    FileSetBuilder* fst_filesetbuilder_new(Context*, BufWriter*);
    void fst_filesetbuilder_insert(Context*, FileSetBuilder*, char*);
    void fst_filesetbuilder_finish(Context*, FileSetBuilder*);

    MemSetBuilder* fst_memsetbuilder_new();
    bool fst_memsetbuilder_insert(Context*, MemSetBuilder*, char*);
    Set* fst_memsetbuilder_finish(Context*, MemSetBuilder*);

    Set* fst_set_open(Context*, char*);
    bool fst_set_contains(Set*, char*);
    size_t fst_set_len(Set*);
    bool fst_set_isdisjoint(Set*, Set*);
    bool fst_set_issubset(Set*, Set*);
    bool fst_set_issuperset(Set*, Set*);
    SetStream* fst_set_stream(Set*);
    SetLevStream* fst_set_levsearch(Set*, Levenshtein*);
    SetOpBuilder* fst_set_make_opbuilder(Set*);
    void fst_set_free(Set*);

    char* fst_set_stream_next(SetStream*);
    void fst_set_stream_free(SetStream*);

    char* fst_set_levstream_next(SetLevStream*);
    void fst_set_levstream_free(SetLevStream*);

    void fst_set_opbuilder_push(SetOpBuilder*, Set*);
    void fst_set_opbuilder_free(SetOpBuilder*);
    SetUnion* fst_set_opbuilder_union(SetOpBuilder*);
    SetIntersection* fst_set_opbuilder_intersection(SetOpBuilder*);
    SetDifference* fst_set_opbuilder_difference(SetOpBuilder*);
    SetSymmetricDifference* fst_set_opbuilder_symmetricdifference(SetOpBuilder*);

    char* fst_set_union_next(SetUnion*);
    void fst_set_union_free(SetUnion*);

    char* fst_set_intersection_next(SetIntersection*);
    void fst_set_intersection_free(SetIntersection*);

    char* fst_set_difference_next(SetDifference*);
    void fst_set_difference_free(SetDifference*);

    char* fst_set_symmetricdifference_next(SetSymmetricDifference*);
    void fst_set_symmetricdifference_free(SetSymmetricDifference*);



    /** ===============================
                     Map
        =============================== **/

    typedef struct {
        char*       key;
        uint64_t    value;
    } MapItem;

    typedef struct FileMapBuilder FileMapBuilder;
    typedef struct MemMapBuilder MemMapBuilder;
    typedef struct Map Map;
    typedef struct MapStream MapStream;
    typedef struct MapLevStream MapLevStream;
    typedef struct MapKeyStream MapKeyStream;
    typedef struct MapValueStream MapValueStream;
    typedef struct MapOpBuilder MapOpBuilder;
    typedef struct MapUnion MapUnion;
    typedef struct MapIntersection MapIntersection;
    typedef struct MapDifference MapDifference;
    typedef struct MapSymmetricDifference MapSymmetricDifference;

    FileMapBuilder* fst_filemapbuilder_new(Context*, BufWriter*);
    bool fst_filemapbuilder_insert(Context*, FileMapBuilder*, char*, uint64_t);
    bool fst_filemapbuilder_finish(Context*, FileMapBuilder*);

    MemMapBuilder* fst_memmapbuilder_new();
    bool fst_memmapbuilder_insert(Context*, MemMapBuilder*, char*, uint64_t);
    Map* fst_memmapbuilder_finish(Context*, MemMapBuilder*);

    Map* fst_map_open(Context*, char*);
    void fst_map_free(Map*);
    uint64_t fst_map_get(Context*, Map*, char*);
    size_t fst_map_len(Map*);
    bool fst_map_contains(Map*, char*);
    MapStream* fst_map_stream(Map*);
    MapKeyStream* fst_map_keys(Map*);
    MapValueStream* fst_map_values(Map*);
    MapLevStream* fst_map_levsearch(Map*, Levenshtein*);

    MapItem* fst_mapstream_next(MapStream*);
    void fst_mapstream_free(MapStream*);
    void fst_mapitem_free(MapItem*);

    char* fst_mapkeys_next(MapKeyStream*);
    void fst_mapkeys_free(MapKeyStream*);

    uint64_t fst_mapvalues_next(Context*, MapValueStream*);
    void fst_mapvalues_free(MapValueStream*);

    MapItem* fst_map_levstream_next(MapLevStream*);
    void fst_map_levstream_free(MapLevStream*);
""")

if __name__ == '__main__':
    ffi.compile()



from contextlib import contextmanager

from .lib import ffi, lib, make_stream_iter, checked_call


class MapBuilder(object):
    def insert(self, val):
        raise NotImplementedError

    def finish(self):
        raise NotImplementedError


class FileMapBuilder(MapBuilder):
    def __init__(self, fpath):
        self._ctx = lib.fst_context_new();
        self._writer_p = checked_call(
            lib.fst_bufwriter_new, self._ctx, fpath.encode('utf8'))
        self._builder_p = checked_call(
            lib.fst_filemapbuilder_new, self._ctx, self._writer_p)

    def insert(self, key, val):
        c_key = ffi.new("char[]", key.encode('utf8'))
        checked_call(lib.fst_filemapbuilder_insert,
                     self._ctx, self._builder_p, c_key, val)

    def finish(self):
        checked_call(lib.fst_filemapbuilder_finish,
                     self._ctx, self._builder_p)
        lib.fst_bufwriter_free(self._writer_p)
        lib.fst_context_free(self._ctx)


class MemMapBuilder(MapBuilder):
    def __init__(self):
        self._ctx = lib.fst_context_new();
        self._ptr = lib.fst_memmapbuilder_new()
        self._map_ptr = None

    def insert(self, key, val):
        c_key = ffi.new("char[]", key.encode('utf8'))
        checked_call(lib.fst_memmapbuilder_insert, self._ctx, self._ptr,
                     c_key, val)

    def finish(self):
        self._map_ptr = checked_call(lib.fst_memmapbuilder_finish,
                                     self._ctx, self._ptr)
        lib.fst_context_free(self._ctx)
        self._ctx = None
        self._ptr = None

    def get_map(self):
        if self._map_ptr is None:
            raise ValueError("The builder has to be finished first.")
        return FstMap(pointer=self._map_ptr)


class OpBuilder(object):
    def __init__(self, map_ptr):
        # NOTE: No need for `ffi.gc`, since the struct will be free'd
        #       once we call union/intersection/difference
        self._ptr = lib.fst_map_make_opbuilder(map_ptr)

    def push(self, map_ptr):
        lib.fst_map_opbuilder_push(self._ptr, map_ptr)

    def union(self):
        stream_ptr = lib.fst_map_opbuilder_union(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_map_union_next,
                                lib.fst_map_union_free)

    def intersection(self):
        stream_ptr = lib.fst_map_opbuilder_intersection(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_map_intersection_next,
                                lib.fst_map_intersection_free)

    def difference(self):
        stream_ptr = lib.fst_map_opbuilder_difference(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_map_difference_next,
                                lib.fst_map_difference_free)

    def symmetric_difference(self):
        stream_ptr = lib.fst_map_opbuilder_symmetricdifference(self._ptr)
        return make_stream_iter(stream_ptr,
                                lib.fst_map_symmetricdifference_next,
                                lib.fst_map_symmetricdifference_free)



class FstMap(object):
    @staticmethod
    @contextmanager
    def build(fpath=None):
        """ Context manager to build a new FST map in a given file.

        Keys must be inserted in lexicographical order.

        See http://burntsushi.net/rustdoc/fst/struct.MapBuilder.html for more
        details.
        """
        if fpath:
            builder = FileMapBuilder(fpath)
        else:
            builder = MemMapBuilder()
        yield builder
        builder.finish()

    @classmethod
    def from_iter(cls, it, fpath=None):
        with cls.build(fpath) as builder:
            for key, val in it:
                builder.insert(key, val)
        if fpath:
            return cls(path=fpath)
        else:
            return builder.get_map()

    def __init__(self, path=None, pointer=None):
        """ Load an FST map from a given file. """
        self._ctx = ffi.gc(lib.fst_context_new(), lib.fst_context_free)
        if path:
            s = checked_call(lib.fst_map_open, self._ctx,
                             ffi.new("char[]", path.encode('utf8')))
        else:
            s = pointer
        self._ptr = ffi.gc(s, lib.fst_map_free)

    def __contains__(self, val):
        return lib.fst_map_contains(
            self._ptr, ffi.new("char[]", val.encode('utf8')))

    def __getitem__(self, key):
        return checked_call(lib.fst_map_get, self._ctx, self._ptr,
                            ffi.new("char[]", key.encode('utf8')))

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return int(lib.fst_map_len(self._ptr))

    def keys(self):
        stream_ptr = lib.fst_map_keys(self._ptr)
        while True:
            c_str = lib.fst_mapkeys_next(stream_ptr)
            if c_str == ffi.NULL:
                break
            yield ffi.string(c_str).decode('utf8')
            lib.fst_string_free(c_str)
        lib.fst_mapkeys_free(stream_ptr)

    def values(self):
        stream_ptr = lib.fst_map_values(self._ptr)
        while True:
            val = lib.fst_mapvalues_next(self._ctx, stream_ptr)
            if val == 0 and self._ctx.has_error:
                break
            yield val
        lib.fst_mapvalues_free(stream_ptr)

    def items(self):
        stream_ptr = lib.fst_map_stream(self._ptr)
        while True:
            itm = lib.fst_mapstream_next(stream_ptr)
            if itm == ffi.NULL:
                break
            c_key = ffi.string(itm.key).decode('utf8')
            yield (c_key, itm.value)
            lib.fst_mapitem_free(itm)
        lib.fst_mapstream_free(stream_ptr)

    def search(self, term, max_dist):
        """ Search the map keys with a Levenshtein automaton.

        :param term:        The search term
        :param max_dist:    The maximum edit distance for search results
        :returns:           Matching (key, value) items in the map
        :rtype:             generator that yields (str, int) tuples
        """
        lev_ptr = checked_call(
            lib.fst_levenshtein_new, self._ctx,
            ffi.new("char[]", term.encode('utf8')), max_dist)
        stream_ptr = lib.fst_map_levsearch(self._ptr, lev_ptr)
        while True:
            itm = lib.fst_map_levstream_next(stream_ptr)
            if itm == ffi.NULL:
                break
            c_key = ffi.string(itm.key).decode('utf8')
            yield (c_key, itm.value)
            lib.fst_mapitem_free(itm)
        lib.fst_map_levstream_free(stream_ptr)
        lib.fst_levenshtein_free(lev_ptr)



import os
import re
import sys
from ._ffi import ffi

class FstError(Exception):
    pass

class TransducerError(FstError):
    pass

class RegexError(FstError):
    pass

class LevenshteinError(FstError):
    pass

class IoError(FstError):
    pass


def find_library():
    libname = "libfstwrapper"
    if sys.platform == 'win32':
        suffix = 'dll'
    elif sys.platform == 'darwin':
        suffix = 'dylib'
    else:
        suffix = 'so'
    cur_dir = os.path.dirname(__file__)
    build_types = ["release", "debug"]
    return os.path.join(cur_dir, "{}.{}".format(libname, suffix))


lib = ffi.dlopen(find_library())

EXCEPTION_MAP = {
    'std::io::Error': OSError,
    'fst::Error': FstError,
    'fst::Error::Fst': TransducerError,
    'fst::Error::Regex': RegexError,
    'fst::Error::Levenshtein': LevenshteinError,
    'fst::Error::Io': IoError,
}


def checked_call(fn, ctx, *args):
    res = fn(ctx, *args)
    if not ctx.has_error:
        return res
    msg = ffi.string(ctx.error_display).decode('utf8').replace('\n', ' ')
    type_str = ffi.string(ctx.error_type).decode('utf8')
    err_type = EXCEPTION_MAP.get(type_str)
    if err_type is FstError:
        desc_str = ffi.string(ctx.error_description).decode('utf8')
        enum_val = re.match(r'(\w+)\(.*?\)', desc_str, re.DOTALL).group(1)
        err_type = EXCEPTION_MAP.get("{}::{}".format(type_str, enum_val))
        if err_type is None:
            msg = "{}: {}".format(enum_val, msg)
    if err_type is None:
        err_type = FstError
    raise err_type(msg)


def make_stream_iter(stream_ptr, next_fn, free_fn):
    while True:
        c_str = next_fn(stream_ptr)
        if c_str == ffi.NULL:
            break
        yield ffi.string(c_str).decode('utf8')
        lib.fst_string_free(c_str)
    free_fn(stream_ptr)



# -*- coding: utf-8 -*-
import pytest

import rust_fst.lib as lib
from rust_fst import Set


TEST_KEYS = [u"möö", u"bar", u"baz", u"foo"]


def do_build(path, keys=TEST_KEYS, sorted_=True):
    with Set.build(path) as builder:
        for key in (sorted(keys) if sorted_ else keys):
            builder.insert(key)


@pytest.fixture
def fst_set(tmpdir):
    fst_path = str(tmpdir.join('test.fst'))
    do_build(fst_path)
    return Set(fst_path)


def test_build(tmpdir):
    fst_path = str(tmpdir.join('test.fst'))
    do_build(fst_path)
    assert True


def test_build_outoforder(tmpdir):
    fst_path = str(tmpdir.join('test.fst'))
    with pytest.raises(lib.TransducerError):
        do_build(fst_path, sorted_=False)


def test_build_baddir():
    fst_path = "/guaranteed-to-not-exist/set.fst"
    with pytest.raises(OSError):
        with Set.build(fst_path) as builder:
            for key in sorted(TEST_KEYS):
                builder.insert(key)

def test_build_memory():
    memset = Set.from_iter(sorted(TEST_KEYS))
    assert len(memset) == 4

def test_load_badfile(tmpdir):
    bad_path = tmpdir.join("bad.fst")
    with bad_path.open('wb') as fp:
        fp.write(b'\xFF'*16)
    with pytest.raises(lib.TransducerError):
        fst_set = Set(str(bad_path))


def test_iter(fst_set):
    stored_keys = list(fst_set)
    assert stored_keys == sorted(TEST_KEYS)


def test_len(fst_set):
    assert len(fst_set) == 4


def test_contains(fst_set):
    for key in TEST_KEYS:
        assert key in fst_set


def test_issubset(tmpdir, fst_set):
    oth_path = tmpdir.join('other.fst')
    do_build(str(oth_path), keys=TEST_KEYS[:-2])
    other_set = Set(str(oth_path))
    assert other_set.issubset(fst_set)
    assert fst_set.issubset(fst_set)


def test_issuperset(tmpdir, fst_set):
    oth_path = tmpdir.join('other.fst')
    do_build(str(oth_path), keys=TEST_KEYS[:-2])
    other_set = Set(str(oth_path))
    assert fst_set.issuperset(other_set)
    assert fst_set.issuperset(fst_set)


def test_isdisjoint(tmpdir, fst_set):
    oth_path = tmpdir.join('other.fst')
    do_build(str(oth_path), keys=[u'ene', u'mene'])
    other_set = Set(str(oth_path))
    assert fst_set.isdisjoint(other_set)
    assert other_set.isdisjoint(fst_set)
    assert not fst_set.isdisjoint(fst_set)
    assert not fst_set.issuperset(other_set)
    assert not fst_set.issubset(other_set)


def test_search(fst_set):
    matches = list(fst_set.search("bam", 1))
    assert matches == [u"bar", u"baz"]


def test_levautomaton_too_big(fst_set):
    with pytest.raises(lib.LevenshteinError):
        next(fst_set.search(u"a"*24, 24))


def test_union():
    a = Set.from_iter(["bar", "foo"])
    b = Set.from_iter(["baz", "foo"])
    assert list(a.union(b)) == ["bar", "baz", "foo"]


def test_difference():
    a = Set.from_iter(["bar", "foo"])
    b = Set.from_iter(["baz", "foo"])
    assert list(a.difference(b)) == ["bar"]


def test_symmetric_difference():
    a = Set.from_iter(["bar", "foo"])
    b = Set.from_iter(["baz", "foo"])
    assert list(a.symmetric_difference(b)) == ["bar", "baz"]


def test_intersection():
    a = Set.from_iter(["bar", "foo"])
    b = Set.from_iter(["baz", "foo"])
    assert list(a.intersection(b)) == ["foo"]



from setuptools import setup


setup(
    name='rust-fst',
    version='0.1',
    author='Johannes Baiter',
    author_email='johannes.baiter@gmail.com',
    description=('Python bindings for the Rust `fst` create, providing sets '
                 'and maps backed by finite state transducers.'),
    license='MIT',
    keywords=['fst', 'rust', 'levenshtein', 'automata', 'transducer',
              'data_structures'],
    url='https://github.com/jbaiter/python-rust-fst',
    setup_requires=['cffi>=1.0.0'],
    install_requires=['cffi>=1.0.0'],
    cffi_modules=['rust_fst/_build_ffi.py:ffi'],
    entry_points={'distutils.setup_keywords': [
        'rust_crates = rust_setuptools:rust_crates'
    ]},
    rust_crates=[('fstwrapper', 'rust_fst')],
    packages=['rust_fst'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Text Processing :: Indexing']
)



# -*- coding: utf-8 -*-
""" Code by Armin Ronacher.

https://github.com/mitsuhiko/rust-setuptools
"""
from __future__ import print_function

import os
import sys
import shutil
import subprocess

from distutils.cmd import Command
from distutils.command.install_lib import install_lib


if sys.platform == 'win32':
    DYNAMIC_LIB_SUFFIX = '.dll'
elif sys.platform == 'darwin':
    DYNAMIC_LIB_SUFFIX = '.dylib'
else:
    DYNAMIC_LIB_SUFFIX = '.so'


class RustBuildCommand(Command):
    description = 'build rust crates into Python extensions'

    user_options = []

    def initialize_options(self):
        for k, v in self.__class__.rust_build_args.items():
            setattr(self, k, v)

    def finalize_options(self):
        pass

    def run(self):
        if self.debug:
            debug_or_release = '--debug'
        else:
            debug_or_release = '--release'

        # Make sure that if pythonXX-sys is used, it builds against the
        # current executing python interpreter.
        bindir = os.path.dirname(sys.executable)
        if sys.platform == 'win32':
            path_sep = ';'
        else:
            path_sep = ':'

        env = dict(os.environ)
        env.update({
            # disables rust's pkg-config seeking for specified packages,
            # which causes pythonXX-sys to fall back to detecting the
            # interpreter from the path.
            'PYTHON_2.7_NO_PKG_CONFIG': '1',
            'PATH':  bindir + path_sep + env.get('PATH', '')
        })

        for crate_path, dest in self.cargo_crates:
            # Execute cargo.
            try:
                toml = os.path.join(crate_path, 'Cargo.toml')
                args = (['cargo', 'build', '--manifest-path', toml,
                    debug_or_release] + list(self.extra_cargo_args or []))
                if not self.quiet:
                    print(' '.join(args), file=sys.stderr)
                output = subprocess.check_output(args, env=env)
            except subprocess.CalledProcessError as e:
                msg = 'cargo failed with code: %d\n%s' % (e.returncode, e.output)
                raise Exception(msg)
            except OSError:
                raise Exception('Unable to execute cargo - this package '
                    'requires rust to be installed and cargo to be on the PATH')

            if not self.quiet:
                print(output, file=sys.stderr)

            # Find the shared library that cargo hopefully produced and copy
            # it into the build directory as if it were produced by
            # build_cext.
            if self.debug:
                suffix = 'debug'
            else:
                suffix = 'release'

            dylib_path = os.path.join(crate_path, 'target/', suffix)

            # Ask build_ext where the shared library would go if it had built it,
            # then copy it there.
            build_ext = self.get_finalized_command('build_ext')

            target = os.path.dirname(build_ext.get_ext_fullpath('x'))
            try:
                os.makedirs(target)
            except OSError:
                pass

            target = os.path.join(target, dest)

            for filename in os.listdir(dylib_path):
                if filename.endswith(DYNAMIC_LIB_SUFFIX):
                    shutil.copy(os.path.join(dylib_path, filename),
                                os.path.join(target, filename))


def build_rust_cmdclass(crates, debug=False,
                        extra_cargo_args=None, quiet=False):
    class _RustBuildCommand(RustBuildCommand):
        rust_build_args = {
            'cargo_crates': crates,
            'debug': debug,
            'extra_cargo_args': extra_cargo_args,
            'quiet': quiet,
        }

    return _RustBuildCommand


def build_install_lib_cmdclass(base=None):
    if base is None:
        base = install_lib
    class _RustInstallLibCommand(base):
        def build(self):
            base.build(self)
            if not self.skip_build:
                self.run_command('build_rust')
    return _RustInstallLibCommand


def rust_crates(dist, attr, value):
    dist.is_pure = lambda: False
    dist.cmdclass['build_rust'] = build_rust_cmdclass(value)
    dist.cmdclass['install_lib'] = build_install_lib_cmdclass(
        dist.cmdclass.get('install_lib'))



from contextlib import contextmanager

from .lib import ffi, lib, make_stream_iter, checked_call


class SetBuilder(object):
    def insert(self, val):
        raise NotImplementedError

    def finish(self):
        raise NotImplementedError


class FileSetBuilder(SetBuilder):
    def __init__(self, fpath):
        self._ctx = lib.fst_context_new();
        self._writer_p = checked_call(
            lib.fst_bufwriter_new, self._ctx, fpath.encode('utf8'))
        self._builder_p = checked_call(
            lib.fst_filesetbuilder_new, self._ctx, self._writer_p)

    def insert(self, val):
        c_str = ffi.new("char[]", val.encode('utf8'))
        checked_call(lib.fst_filesetbuilder_insert,
                     self._ctx, self._builder_p, c_str)

    def finish(self):
        checked_call(lib.fst_filesetbuilder_finish,
                     self._ctx, self._builder_p)
        lib.fst_bufwriter_free(self._writer_p)
        lib.fst_context_free(self._ctx)


class MemSetBuilder(SetBuilder):
    def __init__(self):
        self._ctx = lib.fst_context_new();
        self._ptr = lib.fst_memsetbuilder_new()
        self._set_ptr = None

    def insert(self, val):
        c_str = ffi.new("char[]", val.encode('utf8'))
        checked_call(lib.fst_memsetbuilder_insert, self._ctx, self._ptr, c_str)

    def finish(self):
        self._set_ptr = checked_call(lib.fst_memsetbuilder_finish,
                                     self._ctx, self._ptr)
        lib.fst_context_free(self._ctx)
        self._ctx = None
        self._ptr = None

    def get_set(self):
        if self._set_ptr is None:
            raise ValueError("The builder has to be finished first.")
        return FstSet(pointer=self._set_ptr)


class OpBuilder(object):
    def __init__(self, set_ptr):
        # NOTE: No need for `ffi.gc`, since the struct will be free'd
        #       once we call union/intersection/difference
        self._ptr = lib.fst_set_make_opbuilder(set_ptr)

    def push(self, set_ptr):
        lib.fst_set_opbuilder_push(self._ptr, set_ptr)

    def union(self):
        stream_ptr = lib.fst_set_opbuilder_union(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_set_union_next,
                                lib.fst_set_union_free)

    def intersection(self):
        stream_ptr = lib.fst_set_opbuilder_intersection(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_set_intersection_next,
                                lib.fst_set_intersection_free)

    def difference(self):
        stream_ptr = lib.fst_set_opbuilder_difference(self._ptr)
        return make_stream_iter(stream_ptr, lib.fst_set_difference_next,
                                lib.fst_set_difference_free)

    def symmetric_difference(self):
        stream_ptr = lib.fst_set_opbuilder_symmetricdifference(self._ptr)
        return make_stream_iter(stream_ptr,
                                lib.fst_set_symmetricdifference_next,
                                lib.fst_set_symmetricdifference_free)


class FstSet(object):
    """ An immutable set backed by a finite state transducer stored on disk.

    The interface tries to follow the `frozenset` type as much as possible.
    """
    @staticmethod
    @contextmanager
    def build(fpath=None):
        """ Context manager to build a new FST set in a given file.

        Keys must be inserted in lexicographical order.

        See http://burntsushi.net/rustdoc/fst/struct.SetBuilder.html for more
        details.
        """
        if fpath:
            builder = FileSetBuilder(fpath)
        else:
            builder = MemSetBuilder()
        yield builder
        builder.finish()

    @classmethod
    def from_iter(cls, it, fpath=None):
        with cls.build(fpath) as builder:
            for key in it:
                builder.insert(key)
        if fpath:
            return cls(path=fpath)
        else:
            return builder.get_set()

    def __init__(self, path=None, pointer=None):
        """ Load an FST set from a given file. """
        self._ctx = ffi.gc(lib.fst_context_new(), lib.fst_context_free)
        if path:
            s = checked_call(lib.fst_set_open, self._ctx,
                             ffi.new("char[]", path.encode('utf8')))
        else:
            s = pointer
        self._ptr = ffi.gc(s, lib.fst_set_free)

    def __contains__(self, val):
        return lib.fst_set_contains(
            self._ptr, ffi.new("char[]", val.encode('utf8')))

    def __iter__(self):
        stream_ptr = lib.fst_set_stream(self._ptr)
        while True:
            c_str = lib.fst_set_stream_next(stream_ptr)
            if c_str == ffi.NULL:
                break
            yield ffi.string(c_str).decode('utf8')
            lib.fst_string_free(c_str)
        lib.fst_set_stream_free(stream_ptr)

    def __len__(self):
        return int(lib.fst_set_len(self._ptr))

    def _make_opbuilder(self, *others):
        opbuilder = OpBuilder(self._ptr)
        for oth in others:
            opbuilder.push(oth._ptr)
        return opbuilder

    def union(self, *others):
        return self._make_opbuilder(*others).union()

    def intersection(self, *others):
        return self._make_opbuilder(*others).intersection()

    def difference(self, *others):
        # FIXME: There's a bug in here, it doesn't work as epxected at
        #        the moment
        return self._make_opbuilder(*others).difference()

    def symmetric_difference(self, *others):
        return self._make_opbuilder(*others).symmetric_difference()

    def issubset(self, other):
        return bool(lib.fst_set_issubset(self._ptr, other._ptr))

    def issuperset(self, other):
        return bool(lib.fst_set_issuperset(self._ptr, other._ptr))

    def isdisjoint(self, other):
        return bool(lib.fst_set_isdisjoint(self._ptr, other._ptr))

    def search(self, term, max_dist):
        """ Search the set with a Levenshtein automaton.

        :param term:        The search term
        :param max_dist:    The maximum edit distance for search results
        :returns:           Matching values in the set
        :rtype:             generator that yields str
        """
        lev_ptr = checked_call(
            lib.fst_levenshtein_new, self._ctx,
            ffi.new("char[]", term.encode('utf8')), max_dist)
        stream_ptr = lib.fst_set_levsearch(self._ptr, lev_ptr)
        while True:
            c_str = lib.fst_set_levstream_next(stream_ptr)
            if c_str == ffi.NULL:
                break
            yield ffi.string(c_str).decode('utf8')
            lib.fst_string_free(c_str)
        lib.fst_set_levstream_free(stream_ptr)
        lib.fst_levenshtein_free(lev_ptr)
